/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include <iostream>
#include <cmath>
#include <climits>
#include <fstream>
#include <iomanip>
#include <string.h>
#include "CuttingPlanesEngine.h"    
#include "bendersSubprob.h"
#include "iographs.h"
#include "multiRunsLog.h"
using namespace std;

/*Instance data */
int      ** g;
int      ** gEdge;                  //gives the edge id of an edge in the adjacency matrix
int      n;                         //number of vertices
int      m;                         //number of edges
int*     ii;                        //ii[i] is the first end vertex of edge i
int*     jj;                        //jj[i] is the second end vertex of edge i
double*  b;                         //the demands

/*Algorithm data*/
char*   solution_method_code;       //the alg (with which options) that is run
int     noCutsetConstrIntroRnd = 0; //to introduce randomness, add at start cutset constraints 
#define STDMETH_SMOOTH_ALPHA   0.5  //smoothed solution = last*0.5+current*0.5

/*Stats about performed runs*/
double  bestUb   = INT_MAX;
double  bestUbLP = INT_MAX;         //best fractional ub, only useful for LP solving
int     itersCutPlanes;             //these will be filled by cutPlanes.runCutPlanes(...)
double  tmCutPlanes, finalObj ;
double  totCuts, tmOnlySolve;
long    totalNrCoefs = 0;
int     runId        = 0;
int     failedRuns   = 0;           //This should remain at zero. Otherwise, an error is printed.
int     smoothedCuts           = 0; //number of smoothed cuts used
int     twoCutsUsedSmoothedSep = 0; //unsuccessful smoothed lead to this no of second cuts

/*Log streams*/
char*    texoutfile = NULL;         //If multiple runs, all individual run output can be saved in texoutfile
ofstream texout;

//The code of the methods to be run. Each method is translated into a digit string 
#define CMD_LINE_NAMES 8
char meth[CMD_LINE_NAMES][10] = {"proj","lpproj","proj3", "lpproj3","std" ,"lpstd","lpstd3","std3"};
char code[CMD_LINE_NAMES][10] = {"1100","1000"  ,"1103" , "1003"   ,"0100","0000" ,"0003"  ,"0103"};
char description[CMD_LINE_NAMES][100] = {"projective method IP", "proj meth linear relax LP", "proj meth bandwidth 3", "proj meth LP bandwidth 3", 
                                         "standard   method IP", "std  meth linear relax LP", "std  meth bandwidth 3", "std  meth LP bandwidth 3"};
void expandMethodCodename(char* methCode){
    for(int i=0;i<CMD_LINE_NAMES;i++)
        if((!strcmp(methCode,meth[i]))){
            strcpy(methCode,code[i]);
            clog<<"Method code:"<<methCode<<endl;
        }
    if((methCode[0]<'0')||(methCode[0]>'9')){
        cerr<<"Unknown method code '"<<methCode<<"'. The method code can be one of the following:\n";
        for(int i=0;i<CMD_LINE_NAMES;i++)
            cerr<<setw(10)<<left<<meth[i]<<" :   "<<setw(30)<<description[i]<<code[i]<<endl;
        exit(1);
    }
}

//The following digits are set in the method code, meaning the following:
//digit 0  the method class (0 standard separation, 1 projection) 
//digit 1  discrete only (otherwise fractional LP)
//digit 2  use smoothing for std method (stabilized standard)
//digit 3  consider a bandwidth defined by this digit. set bandwidht =1 if digit 1 not set (or 0) 
int digitIsSet(int digit)
{
    if(strlen(solution_method_code)<(size_t)digit+1)
        return 0;
    char charStringified [2] = {0,0};
    charStringified[0]=solution_method_code[digit];
    return atoi(charStringified);
}
int runStandardSepMethod(){
    return(solution_method_code[0]=='0');
}
int runProjectiveCutPlanes(){
    return(solution_method_code[0]=='1');
}
int useDiscreteOnly(){
    return digitIsSet(1);
}
int useMainSmoothingTechnique(){
    return digitIsSet(2);
}
int useBandwidthDiff1(){
    return digitIsSet(3);
}

void normalizeCut(double* row,double& rHand){
    double maxcoef=row[0];
    for(int i=1;i<m;i++)
        maxcoef=max(maxcoef,row[i]);
    clog<<" following maxcoef will become 10="<<maxcoef<<endl;
    clog<<"normalized cut:";
    for(int i=0;i<m;i++){
        row[i] *= 10.0/maxcoef;
        if(row[i]>0)
            clog<<row[i]<<" ";
    }
    rHand *= 10.0/maxcoef;
    clog<<endl;
}

void firstProjection(double * newCut,double& newRHand, double* xbase)
{
    double*  direction = new double[::m]; //d_i from paper
    double tStar, bTotal = 0;
    for(int i=0;i<n;i++)
        bTotal += b[i];

    for(int i=0;i<(::m);i++){
        xbase[i] = bTotal;
        direction[i] = -xbase[i];
    }

    tStar = newProjection(xbase,direction, newCut,newRHand);
    for(int i=0;i<(::m);i++){
        xbase[i] = xbase[i] + 0.2*tStar*direction[i];
    }
    normalizeCut(newCut,newRHand);
    delete[] direction;
}

double cutBendersProjective(double*opt, double * newCut,double&newRHand, int it, double tm)
{
    static double*  xbase     = new double[::m]; //xbase is interior point x_i from the paper
    static double*  direction = new double[::m]; //d_i from paper
    double tStar  ;
    if(it==0){
        firstProjection(newCut,newRHand, xbase);
        return  -1;
    }

    //Code to print solutions and investigate the bang bang effects. Also put 1
    //instead of stepLen in the xbase update below (but also above for 1st iter)
    //static int printed=0;
    //    if(printed%10==0){
    //    double sum=0;
    //    for(int i=0;i<15;i++){
    //        cout<<setw(5)<<left<<setprecision((xbase[i]>1000)?4:(xbase[i]>0.1?3:2))<<xbase[i]<<" ";
    //        sum+=xbase[i];
    //    }
    //    cout<<"TT"<<sum<<endl;
    //}
    //printed++; if(printed==60) exit(1);


    //Round input (solve num pb) and calc direction
    for(int i=0;i<(::m);i++){
        if(opt[i]<EPSILON)
            opt[i] = 0;
        direction[i] = opt[i] - xbase[i]; //xbase is interior point x_i from the paper
    }
    clog<<"\ncurrent opt sol x=";
    for(int i=0;i<(::m);i++)
        clog<<opt[i]<<",";
    clog<<endl;

    //Solve projection
    int checkHitPoint = 0;
    if(useDiscreteOnly())                 //On MILP, the integer master is more expensive
        checkHitPoint = 1;                //than any hit point feasibility checking
    
    tStar = newProjection(xbase,direction, newCut,newRHand, checkHitPoint);

    //Calculate violation, how much the new cut is violated by current sol opt
    double violationDegree = 0;
    for(int i=0;i<m;i++)
        violationDegree += newCut[i]*opt[i];
    violationDegree-= newRHand;
    clog<<"violation degree="<<violationDegree<<endl;
    if(violationDegree>=-EPSILON)
        clog<<"FEASIB SOL|";

    //Determine up bound and log info
    double ub     = 0;
    double ub_int = 0;
    for(int i=0;i<(::m);i++){
        ub     += (xbase[i] + tStar*direction[i]);
        ub_int += ((int)ceil(xbase[i] + tStar*direction[i]-EPSILON));
    }
    bestUb = min(bestUb,ub_int);
    clog<<"IT "<<setw(3)<<it<<" TM"<<setw(5)<<tm<<" INTUB="<<ub_int
        <<" INTUB BST="<<bestUb<<endl;
    if(ceil(lowerBound-EPSILON)>=ceil(bestUb-EPSILON)) //if int gap = 0 |since goal=B&B 
       return INT_MAX;                                 //problem solved |even for LP solve
    if(!useDiscreteOnly()){
        bestUbLP = min(bestUbLP,ub);
        clog<<" LPUB="<<setw(9)<<ub<<"BST LPUB="<<bestUbLP<<endl;
        //since goal=B&B, a zero gab on rounded bounds means ceil[opt] is found already
        if(ceil(lowerBound-EPSILON)>=ceil(bestUbLP-EPSILON))
           return INT_MAX;                                
    }
    if(runId>0)                                 //tmp msg in multi run mode
        cerr<<setw(4)<<it<<" "<<setw(5)<<tm<<setw(9)<<lowerBound<<setw(5)<<ub_int<<setw(9)<<ub<<endl;
    //Determine new xbase
    double stepLen = 0.2;
    if((it>100)&&(useDiscreteOnly()))
        stepLen = 0.4;

    //stepLen = 0.90;//could use checkHitPoint=1 when stepLen>=.9, as num pb arise

    for(int i=0;i<(::m);i++){
        xbase[i]      = xbase[i] + stepLen*tStar*direction[i];
        //if(xbase[i]<EPSILON)
        //    xbase[i] = 0;
    }

    //Normalize the new cut (useful to remove numerical problems) and return
    normalizeCut(newCut,newRHand);
    return violationDegree;
}
//returns a violation degree of yyy, ie., x^Ta-b, where x=yyy, a=newRow, b=newRightHand
double cutBendersSubprob (const int nrVars, double*yyy, double * newRow,double&newRightHand,
                int it, double tm,
                double ** newColMore,double*newRightHandMore,
                int&newMore, int maxMoreConstr)
{ 
    if(runProjectiveCutPlanes()){
        newMore = 0;
        return cutBendersProjective(yyy,newRow,newRightHand,it,tm);
    }
    clog<<"-----START STD BENDERS SEPARATOR------\n";
    newMore = 0;
    clog<<"Input cutBendersSubprob i(endvtx1,endvtx2):y[i]....=";
    for(int i=0;i<m;i++){
        if(yyy[i]!=0){
            clog<<i;
            clog<<"("<<ii[i]<<","<<jj[i]<<"):"<<yyy[i]<<", ";
        }
    }
    clog<<endl;
    double violationDegree = 0;
    if(useMainSmoothingTechnique()){
        static double* yLast     = new double[m];
        static double* ySmoothed = new double[m];
        if(smoothedCuts==0){
            newCut(yyy,newRow, newRightHand);
        }else{
            for(int i=0;i<m;i++){
                ySmoothed[i] = yLast[i] * STDMETH_SMOOTH_ALPHA + 
                               yyy[i]   * (1-STDMETH_SMOOTH_ALPHA);
                 if(ySmoothed[i]<EPSILON)
                     ySmoothed[i] = 0;
            }
            newCut(ySmoothed,newRow, newRightHand);
            double newSmoothedRowValue = 0;
            for(int i=0;i<m;i++)
                if(yyy[i]>EPSILON)
                    newSmoothedRowValue += newRow[i]*yyy[i];
            if(newSmoothedRowValue<newRightHand-EPSILON){
                clog<<"One smoothed constraint will be enough\n";
            }
            else{
                clog<<"Two constraints will be returned (smoothed one not enough)\n";
                twoCutsUsedSmoothedSep++;
                assert(newMore<maxMoreConstr);
                for(int i=0;i<m;i++){
                    newColMore[newMore][i]     = newRow[i];
                    newRightHandMore[newMore] = newRightHand;
                }
                newMore++;
                newCut(yyy,newRow, newRightHand);
            }

       }
       smoothedCuts++;
       for(int ii=0;ii<m;ii++)
            yLast[ii] = yyy[ii]; 
    }
    else{//no smoothing asked
        newCut(yyy,newRow, newRightHand);
        clog<<"plotlb"<<"("<<it<<","<<lowerBound<<")"<<endl;
    }

    for(int i=0;i<m;i++)
        violationDegree += newRow[i]*yyy[i];
    violationDegree-= newRightHand;
    clog<<"Generated constraint has a right hand value newRightHand="<<newRightHand<<endl;
    clog<<"Final message from cutBendersSubprob is violationDegree="<<violationDegree<<endl;
    clog<<"-----END BENDERS SEPARATOR------\n";
    return violationDegree;
}

void addInitialCuts(CuttingPlanesEngine &cutPlanes, int nrCutsetConstr){
    int * connectedComponent = new int[n];
    double*cutSetConstraint = new double[m];
    double demandTotal;
    clog<<"Constraints regarding runId="<<runId<<endl;
    unsigned int rndSeed = time(NULL);
    clog<<"rndSeed="<<rndSeed<<endl;
    if(texoutOpen()){          //write info to file provided by -fTEXFILENAME option, if it exists
        texout<<"RNDSEED="<<rndSeed<<" and nrInitialConstraints="
                <<noCutsetConstrIntroRnd<<" at run "<<runId
                <<" on "<<getGraphLabelForStats()
                <<" with m="<<m<<" and n="<<n
                <<endl;
        texout.close();
    }
    srand(rndSeed);
    clog<<"I will add "<<nrCutsetConstr<<"  initial random cutset constraints\n";
    for(int loop=0;loop<nrCutsetConstr;loop++){
        for(int i=0;i<n;i++)
             connectedComponent[i] = 1;
        for(int ne=0;ne<n/10;ne++)                  //random edges
             connectedComponent[rand()%(n-1)]=0;
        clog<<"Generated the "<<(1+loop)<<"th random cut set...";
        if(newCutSet(connectedComponent,cutSetConstraint,demandTotal))
             cutPlanes.modelAddCut(cutSetConstraint,demandTotal);
    }
    clog<<"All done\n";
    delete[] connectedComponent;
    delete[] cutSetConstraint;
}

void checkPrintDegrees(ostream&localout){
    int* degrees = new int[n];
    for(int v=0;v<n;v++)
        degrees[v]=0;
    for(int e=0;e<m;e++){
        degrees[ii[e]]++;
        degrees[jj[e]]++;
    }
    localout<<getGraphLabelForStats()<<"Degrees:";
    for(int v=0;v<n;v++)
        localout<<degrees[v]<<" ";
    localout<<endl;
    delete[] degrees;
}
int checkGraphConnectivity(ostream&localout){
    static int alreadyChecked = 0;
    if(alreadyChecked)          //do not do it twice one the same graph
        return 1;
    int* connectComp = new int[n];
    for(int i=0;i<n;i++)
        connectComp[i] = 0;
    connectComp[n-1] = 1;
    fillConnCompUpToDepth(connectComp,NULL,n-1,n);
    //clog<<"Connected component:";
    //for(int i=0;i<n;i++)
    //    clog<<i<<":"<<connectComp[i]<<" ";
    //clog<<endl;
    for(int i=0;i<n;i++)
        if(!connectComp[i])
            if(b[i]>0){
                localout<<"Vertex "<<i<<" has demand "<<b[i]<<" but it is not connected to the source";
                localout<<"\nThe instance "<<getGraphLabelForStats()<<"m="<<m<<" n="<<n<<"is infeasible. I return failure.\n";
                cerr<<"Vertex "<<i<<" has demand "<<b[i]<<" but it is not connected to the source";
                cerr<<"\nThe instance "<<getGraphLabelForStats()<<"m="<<m<<" n="<<n<<"is infeasible. I return failure.\n";
                return 0;
            }
    alreadyChecked = 1;
    delete[] connectComp;
    checkPrintDegrees(clog);
    return 1;//ok
}
int checkGraphConnectivityWithLog(){
    if(texoutOpen()){
        int retValue = checkGraphConnectivity(texout);
        texout<<endl;
        texout.close();
        return retValue;
    }
    return checkGraphConnectivity(clog);
}

int oneRun()
{                                   
    CuttingPlanesEngine cutPlanes(m,cutBendersSubprob,10);   //maximum 10 additional constraints can be returned by subprobBenders
    if(runId==0){                                            //no multiple runs
        if(clog.fail())     
           cout<<"clog is disabled  "<<endl;
        cutPlanes.activateLog();
    }
    //cutPlanes.printProgressMsg(cout);
    //cutPlanes.alwaysTurnIntegerInTheEnd();
    //cutPlanes.activateLog();
    //cutPlanes.setVarBounds(0,10000);
    double * coefsObj;
    coefsObj = new double[m];
    for(int i=0;i<m;i++)
        coefsObj[i] = 1;
    cutPlanes.setObjCoefs(coefsObj);
    if(useDiscreteOnly())
        cutPlanes.turnAllVarsInteger();

    if(noCutsetConstrIntroRnd>0)
        addInitialCuts(cutPlanes,noCutsetConstrIntroRnd);

    clog<<"=================================================================================="<<endl;
    clog<<"Starting constraint generator using the Benders separator indicated by constructor"<<endl;

    if(cutPlanes.runCutPlanes(100000, 150000, itersCutPlanes, tmCutPlanes)==EXIT_FAILURE){
        cerr<<"Exiting Cutting Planes Engine with ERROR. Maybe not enough time or iterations to optimize to the full (or error? Check error messages above)\n";
	    if(runId==0){
		    cerr<<"The problematic run is runId="<<runId
                <<" on "<<getGraphLabelForStats()
                <<" with m="<<m<<" and n="<<n<<endl;
        }
        if(texoutOpen()){
            texout<<"Exiting Cutting Planes Engine with ERROR. Maybe not enough time or iterations to optimize to the full (or error? Check error messages)\n";
            texout<<"Time elapsed "<<tmCutPlanes<<" iters="<<itersCutPlanes<<endl;
		    texout<<"The problematic run is runId="<<runId
                <<" on "<<getGraphLabelForStats()
                <<" with m="<<m<<" and n="<<n<<endl;
            texout.close();
        }
        finalObj = cutPlanes.getObjVal();
        clog<<"Final obj val (with timeout or error)="<<finalObj<<" obtained after "<<itersCutPlanes<<" iters of CutPlanes and "<<tmCutPlanes<<" secs.\n";
        totalNrCoefs+=cutPlanes.getTotalNrCoefs();
        return 0;
    }
    finalObj   = cutPlanes.getObjVal();
    totCuts    = cutPlanes.getNbCuts();
    tmOnlySolve= cutPlanes.getTmOnlySolve();

    //Print final information
    clog<<"Final obj val="<<finalObj<<" obtained after "<<itersCutPlanes<<" iters of Cutting Planes and "<<tmCutPlanes<<" secs. Final solution below under the form (edgEndVtx1,edgEndVtx2):ysol [edg]=value\n";
    double*y = new double[m]; 
    cutPlanes.getPrimals(y); 
    for(int i=0;i<m;i++){
        if(y[i]!=0){
            clog<<"("<<ii[i]<<","<<jj[i]<<"): ";
            clog<<"ysol ["<<i<<"]="<<y[i]<<"; ";
        }
    }
    clog<<endl;

    totalNrCoefs+=cutPlanes.getTotalNrCoefs();

    if(runId==0){
        latexLineForLastRun(cout);
        lineBoxForLastRun(cout);
    }else{
        latexLineForLastRun(clog);
        lineBoxForLastRun(clog);
    }
    if(texoutOpen()){
        latexLineForLastRun(texout);
        texout.close();
    }

    return 1;//ok
}
/*-----------------------+------------------------------------+--------------------------
                         |PARSE CMD LINE AND LAUNCH ALGORITHMS|
                         +-----------------------------------*/

void multipleRuns(char* nrRunsStr)
{
    int runs;
    runs = atoi(nrRunsStr);
    clog<<"I will execute the following nr of runs:"<<runs<<endl;
    for(runId = 1; runId<=runs;runId++){
        bestUb   = INT_MAX;
        bestUbLP = INT_MAX;
        smoothedCuts = 0;
        twoCutsUsedSmoothedSep = 0;
        
        clog<<"Run "<<runId<<" on "<<getGraphLabelForStats()<<endl;
        if(oneRun())
            addToStats();
    }
}
void checkAllParamsAreOk(int& argc, char** argv)
{
    while(argv[argc-1][0]=='-'){
        int argc_start = argc;
        if(argv[argc-1][1]=='s') {
            clog.setstate(ios_base::failbit);
            argc--;
        }
        if(argv[argc-1][1]=='r') {
            clog<<"I will add 10 cut set constraints to randomize algorithms\n";
            noCutsetConstrIntroRnd = 10;
            argc--;
        }
        if(argv[argc-1][1]=='f'){    
            texoutfile = new char[strlen(argv[argc-1])];
            strcpy(texoutfile,argv[argc-1]+2);
            clog<<"Writing tex results of each run in "<<texoutfile<<endl;
            argc--;
        }
        if(argc_start==argc){
            cerr<<"I could not parse option '"<<argv[argc-1]<<"' The options I accept"
                  " are: -fTEXLOGFILE (no spaces), -s[ilent], -r[nd]\n";
            exit(EXIT_FAILURE);
        }
    }

    for(int i=3;i<argc;i++)
         if(!((argv[i][0]>='0')&&(argv[i][0]<='9'))){    
           cerr<<"Argument "<<i<<" [runs?] does not start with a digit or with '-'. Exit\n";
           exit(EXIT_FAILURE);
         }


    if((argc<3)||(argc>6)){
        cerr<<endl;
        cerr<<"Usage1: ./main method(\"std\", \"proj\", ...) instance [-s[ilent]] [-fTEXOUTFILE] [-r[nd]] [runs] \n"
              "Usage2: ./main method(\"std\", \"proj\", ...) m(edges) n(vtx) rndseed [-s[ilent]] -[-fTEXOUTFILE] [-r[nd]] [runs] \n";
        cerr<<"                Use -s to disable writing to log, -fTEX to write tex stats to TEX file.\n"
              "                Use -r[nd] to randomize algo by inserting 10 cut set constraints at start\n";        
        cerr<<"                If you provide an optional number of runs [runs], the program will only print the final statistics and nothing else\n";
        cerr<<"                If you do not provide a number of runs, it will print the progress messages of the sub-problems and the Cutting Planes \n";
        cerr<<"The method field can be one of the following:\n";
        for(int i=0;i<CMD_LINE_NAMES;i++)
            cerr<<setw(8)<<left<<meth[i]<<" :  "<<setw(27)<<description[i]<<"code "<<code[i]<<endl;
        cerr<<"You can also directly write the code. Set the third\nbit to use stabilization (smoothing) in std method.\n";
        exit(EXIT_FAILURE);
    }
}

int main(int argc, char** argv)
{
    #ifdef NDEBUG
    clog.setstate(ios_base::failbit);
    #endif
    checkAllParamsAreOk(argc,argv);
    solution_method_code = new char[50];
    strncpy(solution_method_code,argv[1],45);
    expandMethodCodename(solution_method_code);
    setBandwidth(1);
    if(useBandwidthDiff1())
        setBandwidth(useBandwidthDiff1());

    if(argc>=5){//usage 1
        m         = atoi(argv[2]);
        n         = atoi(argv[3]);
        int rndSd = atoi(argv[4]);

        genRndGraphAndDemandsOrLoadFromAssocFileWhenItExists(rndSd);
        if(argc==5){
            if(!checkGraphConnectivity(cout))
                exit(1);
            clog<<"One run:\n";
            oneRun();
            return 0;
        }
        if(argc==6){
            if(!checkGraphConnectivityWithLog())
                exit(1);
            multipleRuns(argv[5]);
        }
    }
    if(argc<5){        
        loadGraphFromFile(argv[2]);
        if(argc==3){
                if(!checkGraphConnectivity(cout))
                    exit(1);
                clog<<"One run on "<<argv[2]<<":\n";
                oneRun();
                return 0;
        }
        if(argc==4){
            if(!checkGraphConnectivityWithLog())
                exit(1);
            multipleRuns(argv[3]);
        }
    }

    #ifdef NDEBUG
    clog.clear();
    #endif
    printStats();
}
