/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include "multiRunsLog.h"
#include "statManager.h"
#include <climits>
#include <cmath>
#include <iomanip>
#include <iostream>
using namespace std;

int    runStandardSepMethod();
int    useMainSmoothingTechnique();
int    useDiscreteOnly();
double getBandwidth();
char*  getGraphLabelForStats();

int texoutOpen()
{
    if(texoutfile==NULL)
        return 0;
    texout.open(texoutfile, std::ios::out | std::ios::app);
    if (texout.fail()){
        cerr<<"Could not write to"<<texoutfile<<endl;
        return 0;
    }
    texout.exceptions(texout.exceptions() | std::ios::failbit | std::ifstream::badbit);
    return 1;
}

void lineBoxForLastRun(ostream&localout)
{
    localout<<endl<<endl;
    localout<<"                               FINAL RESULTS\n";
    double bMaxDem = -1;
    for(int i=0;i<n;i++)
        bMaxDem = max(bMaxDem,b[i]);
    if(getBandwidth()>1)    
        bMaxDem*=getBandwidth();
    localout<<"+-------------------------------------------------------------------------";
    if(!useDiscreteOnly())
        localout<<"---------";
    localout<<"+\n";
    localout<<"|method code|edges| vtx |max dem|bndWd|| OPT  |";
    if(!useDiscreteOnly())                         //lp fractional case
                localout<<" INT UB |";
    localout<<" iters |CPU tm|";
    if(useDiscreteOnly())
        localout<<"restr ILP tm|";
    else
        localout<<"restr LP  tm|";
    localout<<endl;
    localout<<"|"<<left<<setw(11)<<solution_method_code;
    localout<<"|";
    int precisionReportTm = 1;
    if(tmCutPlanes>=1)
        precisionReportTm = 2;
    if(tmCutPlanes>=10)
        precisionReportTm = 3;
    if(tmCutPlanes>=1000)
        precisionReportTm = 4;
    if(tmCutPlanes>=10000)
        precisionReportTm = 5;

    localout<<setprecision(5)<<left;
    if(finalObj==INT_MAX){
        localout<<setw(16)<<setw(5)<<m<<"|"<<setw(5)<<n<<"|"<<setw(7)<<bMaxDem<<"|"<<setw(5)<<getBandwidth()<<"||"<<setw(6)<<"Failed! "
            <<"You set a time out via cutPlanes.setTimeoutSolve()?   ";
    }else{
        localout<<setw(16)<<setw(5)<<m<<"|"<<setw(5)<<n<<"|"<<setw(7)<<bMaxDem<<"|"<<setw(5)<<getBandwidth()<<"||"<<setw(6)<<setprecision(5)<<finalObj<<"|";
        if(!useDiscreteOnly()){     //lp fractional case, used often for bandwidth>0 (hardest instances)
            if(runStandardSepMethod())
                localout<<"  ----  |";
            else
                localout<<" "<<setw(7)<<bestUb<<"|";
        }
    }
    localout<<setw(7)<<itersCutPlanes <<"|"<<setw(6)<<setprecision(precisionReportTm)<<tmCutPlanes<<"|";
    double percentOnlySolve = tmOnlySolve/tmCutPlanes*100;
    localout<<"    "<<setw(4)<<left<<setprecision(3)<<percentOnlySolve<<"%   |";
    localout<<endl;
    localout<<"+-------------------------------------------------------------------------";
    if(!useDiscreteOnly())
        localout<<"---------";
    localout<<"+\n";
    localout<<endl;
}
void latexLineForLastRun(ostream&localout){
    localout<<""<<left<<setw(11)<<solution_method_code;
    int precisionReportTm = 1;
    if(tmCutPlanes>=1)
        precisionReportTm = 2;
    if(tmCutPlanes>=10)
        precisionReportTm = 3;
    if(tmCutPlanes>=1000)
        precisionReportTm = 4;
    if(tmCutPlanes>=10000)
        precisionReportTm = 5;

    localout<<setprecision(5)<<left;
    if(finalObj==INT_MAX){
        localout<<setw(16)<<getGraphLabelForStats()<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<"Failed! "
            <<"You set a time out via cutPlanes.setTimeoutSolve()?   ";
    }else{
        localout<<setw(16)<<getGraphLabelForStats()<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<setprecision(5)<<finalObj<<"&";
        if(!useDiscreteOnly()){     //lp fractional case, used often for bandwidth>0 (hardest instances)
            if(runStandardSepMethod())
                localout<<"--      &";
            else
                localout<<"ub"<<setw(6)<<bestUb<<"&";
        }
    }
    localout<<setw(5)<<itersCutPlanes <<"&"<<setw(6)<<setprecision(precisionReportTm)<<tmCutPlanes<<"&";
    double percentOnlySolve = tmOnlySolve/tmCutPlanes*100;
    localout<<setw(4)<<left<<setprecision(3)<<percentOnlySolve<<"\\%";
    localout<<"\\\\%";
    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique())
            localout<<"SmoothedOneEnghTot:"<<smoothedCuts-twoCutsUsedSmoothedSep <<"/" <<smoothedCuts<<",";
    }

    localout<<setprecision(5);
    localout<<"NbCuts"<<setw(5)<<totCuts
          <<" TmOnlySolve "<<setw(7)<<tmOnlySolve<<endl;
}


/*-----------------------+------------------------------------+--------------------------
                         |   ROUTINES TO GENERATE STATISTICS  |
                         +-----------------------------------*/
#define EPSILON 1e-06
double   finalObjFirstRun    = -1;
int      noFinalObjDiffFirst =  0;
statManager itersCutPlanes_stat, tmCutPlanes_stat;
statManager totCuts_stat, tmOnlySolve_stats;
statManager totalSmoothed_stat,totalOneEnoughForSmoothed_stat;
double bestUbAllStats = INT_MAX;
void addToStats()
{
    if(finalObjFirstRun==-1)
        finalObjFirstRun = finalObj;
    if(abs(finalObjFirstRun-finalObj)>EPSILON){
        cerr<<"&STAT-ATTENTION: found finalObj="<<finalObj<<" at run "<<runId<<" but the first run found finalObjFirstRun"<<finalObjFirstRun<<endl;
        noFinalObjDiffFirst++;
    }
    if(finalObj==INT_MAX){
        cerr<<"&STAT-ATTENTION: failed to finish optimization\n";
        failedRuns++;
        return ;
    }
    itersCutPlanes_stat.addValue(itersCutPlanes);
    tmCutPlanes_stat.addValue(tmCutPlanes);

    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique()){
            totalSmoothed_stat.addValue(smoothedCuts);
            totalOneEnoughForSmoothed_stat.addValue(smoothedCuts-twoCutsUsedSmoothedSep);
        }
    }
    totCuts_stat.addValue(totCuts);
    tmOnlySolve_stats.addValue(tmOnlySolve);
    if(!runStandardSepMethod())
        if(!useDiscreteOnly())
            bestUbAllStats = min(bestUbAllStats,bestUb);
}

void printStats(){
    if(getBandwidth()>1)
        cout<<"BND"<<getBandwidth();
    if(noFinalObjDiffFirst>0)
        cout<<"?"<<noFinalObjDiffFirst;
    else
        cout<<"  ";
    cout<<"STAT"<<setw(2)<<itersCutPlanes_stat.getNrValues()<<" ";;
    cout<<setprecision(5)<<left;
    cout<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<finalObjFirstRun<<"&"<<setw(15)<<getGraphLabelForStats()<<"";
    cout<<left<<setw(10)<<solution_method_code<<" ";

    if(!runStandardSepMethod())
        if(!useDiscreteOnly())
            cout<<"BSTUB "<<setw(5)<<bestUbAllStats;
    cout<<"ITERS "<<itersCutPlanes_stat.getStats()<<"TIME ";
    cout<<tmCutPlanes_stat.getStats();
    int percentMaster = 1000.0*tmOnlySolve_stats.getAvg()/tmCutPlanes_stat.getAvg();
    if(percentMaster>=1000)  
        percentMaster = 999;
    if(percentMaster<1)
        percentMaster = 1;
    cout<<setw(4)<<setprecision(3)<<right<<(double)percentMaster/10<<"\\% ";
    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique())
            cout<<"SmoothedOneEnghTot:"<<totalOneEnoughForSmoothed_stat.getStats()
                 <<"\\\\%/"<<totalSmoothed_stat.getStats()<<",";
        for(int loop=0;loop<=235;loop++)    //put padding
            cout<<" ";
    }
    cout<<" Total Coefs "<<totalNrCoefs<<"";
    cout<<" |NbCuts"<<totCuts_stat.getStats()
          <<" |TmOnlySolve "<<tmOnlySolve_stats.getStats();

    if(failedRuns>0)
        cout<<"\n&STAT-ATTENTION: the following number of runs failed:"<<failedRuns;
    cout<<endl;
}

